/*

DESCRIPTION:
    Main program file for TS50 inverter which is a modified inv48 R&D unit.
    Intended to deliver maximum of 50W with buck stage pre-regulating the
    input to inverter since it's functioning as a DC transformer (no regulation).

SYNTAX:  n/a

KNOWN BUGS:  n/a

COPYRIGHT:  (c) 2015, 2018, Brian Cornell, all rights reserved.

DEPENDENCIES:
    main.h - configuration constants for main.c
    usart.c - RS232 service module
    usart.h - RS232 service module

VERSION:  1.0, development release

MODIFICATION HISTORY:

    25 Mar 18, BWC, initial development release.
    

*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>
#include <binary.h>
#include "config.h"
#include "usart.h"
#include "main.h"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                              Data objects
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
invcf Config;                               //Configuration data
bit cmdGood;                                //TRUE if command successful
bit statOpen;                               //TRUE if running in open loop
bit statDigital;                            //TRUE if using digital control loop
bit statMonitor;                            //TRUE if console monitoring on
bit incr;                                   //TRUE if digital loop should increase DC
bit prnResult;                              //TRUE if detects console access


const unsigned char nvData[FLASHROWSZ] = {S_BL_AUTOD, STS_UND, S_DC_RUN,\
 S_DC_RUN_P, TRUE, CFG_DAC_5_VH, CFG_DAC_5_VL, FALSE};
unsigned char ramData[FLASHROWSZ],          //RAM copy of NVM config data
        monTimer;                           //Used by monitor feature & maintained by T2 ISR



/*Function main.  Enter infinite loop and wait for Timer 2 interrupt for completion
 * of PWM period.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
    char *CmdBuffer;
    unsigned char cmdCount, avgCount = S_IN_SAMPLES, ovcCount = 0, ovpCount = 0,
            uvlCount = 0;
    unsigned int delta, icsAvgSample = 0, invAvgSample;
    volts icsSample, invSample;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Boot MCU, initialize peripherals, and get ready for operation.  Afterwards
     * take a pause for environment to settle.
     * 
     * Take a one-time sample of inverter voltage so autostart & under-volt
     * lockout processes work correctly.  If not done, the minimum voltage that
     * the inverter will start at is S_VL_INV_RSTART which is not desired
     * behavior.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Initialize();
    __delay_ms(S_TM_INT);
    ReadVoltage(PIN_I_IVDCPV, S_RT_INV, &Config.inv);

    while (TRUE)
    {
        CLRWDT();                           //Clear watchdog timer

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Scan operating environment.
         * 
         * Various system voltages are read first.  Those that require averaging
         * are stored in temporary objects for summing & subsequent averaging.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReadVoltage(PIN_I_IVDCPV, S_RT_INV, &invSample);        //Temp storage for averaging
        ReadVoltage(PIN_I_IVCSSH, S_RT_ICS, &icsSample);        //Temp storage for averaging
        ReadVoltage(PIN_I_INVTMP, S_RT_TMP, &Config.tmp);
        ReadVoltage(PIN_I_IVVCCV, S_RT_SEN, &Config.sen);

        getCOGtime();                                           //Fetch duty cycle


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Calculate average of measured samples that require averaging.  On 
         * each loop sum the measured value and, when sample count is reached
         * average the value.
         * 
         * Then perform a 1st order calculation of primary amperes & power.  For 
         * reasonable accuracy 95% of the value represented by Config.ics.vmv 
         * must be used since this is the fixed DC of the inverter.
         * 
         * Note:  when the inverter is not running (specifically the COG is off),
         * the current S&H circuit voltage floats and will indicate large 
         * current values.  If not cleared this will interfere with the run-time
         * checks immediately after start before a new sample set is taken.  To
         * prevent this the ics sample is cleared when the status is not RUN.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        invAvgSample += invSample.adc;                          //Sum raw value
        icsAvgSample += icsSample.adc;                          //Sum raw value
        if (++avgCount > S_LM_SAMPLES)
        {
            //Hack to clear noise readings from S&H circuit when COG is off
            if (Config.sts != STS_RUN) icsAvgSample = 0;

            //Calculate averages
            CalcAvg(invAvgSample, S_LM_SAMPLES, S_RT_INV, &Config.inv);
            CalcAvg(icsAvgSample, S_LM_SAMPLES, S_RT_ICS, &Config.ics);

            //Convert current volts value to amps & calculate power
            Config.amps = ((Config.ics.vmv - ((Config.ics.vmv * 25) / 100)) / S_VL_MV_PER_AMP);
            Config.pwr = ((Config.amps * (Config.inv.vmv / 100)) / 10);
            avgCount = S_IN_SAMPLES;                            //Reset counter
            icsAvgSample = 0;                                   //Clear sample
            invAvgSample = 0;                                   //Clear sample
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * RUN-TIME CODE BLOCKS:  digital control loop, run-time diagnostic
         * checks.  Grouped together under status RUN conditional for efficiency
         * but must not interfere with one another's processing priority.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts == STS_RUN)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Digital control loop; simple P-I loop.
             * 
             * Experimental and not for production.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if ((statDigital) && !(statOpen))
            {
                if (Config.sen.vmv > Config.ref.vmv)
                {
                    incr = FALSE;
                    delta = Config.sen.vmv - Config.ref.vmv;
                }
                else
                {
                    incr = TRUE;
                    delta = Config.ref.vmv - Config.sen.vmv;
                }
                if (delta / 100)
                {
                    if (incr)
                    {
                        if (PWM5DCL < S_DC_MAX) ++PWM5DCL;
                    }
                    else
                    {
                        if (PWM5DCL > S_DC_MIN) --PWM5DCL;
                    }
                    PWM5LDCONbits.LDA = TRUE;
                }
            }


            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * RUN-TIME CHECKS:  relay, current, voltage, temperature, load, sine.
             * 
             * Waterfall loop:  first error sets status & breaks loop.  Hence, the
             * most important / time critical checks should be done first.  Safety
             * shutdown is performed if any check evaluates TRUE.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            while (TRUE)
            {
                //Relay.
                if (!(PIN_I_RELAY))
                {
                    Config.sts = STS_RLY_CMD;
                    break;
                }

                //Over-current.  Simple P-D loop.
                if (Config.cogc < Config.oldc) ++ovcCount; else if (ovcCount) --ovcCount;
                if (ovcCount > S_LM_OVC_HYS)
                {
                    ovcCount = 0;
                    Config.sts = STS_OVC_LCKO;
                    break;
                }

                //Low voltage
                if (Config.inv.vmv < S_VL_INV_MIN) ++uvlCount; else if (uvlCount) --uvlCount;
                if (uvlCount > S_LM_UVL_HYS)
                {
                    uvlCount = 0;
                    Config.sts = STS_BUS_UVLO;
                    break;
                }

                //Temperature
                if (Config.tmp.vmv < S_VL_TMP_OVH)
                {
                    Config.sts = STS_HIGH_TEMP;
                    break;
                }

                //Overload (power)
                if (Config.pwr > S_LM_PWR_MAX) ++ovpCount; else if (ovpCount) --ovpCount;
                if (ovpCount > S_LM_OVP_HYS)
                {
                    ovpCount = 0;
                    Config.sts = STS_OVERLOAD;
                    break;
                }

                //Sine.
                if ((PIN_O_SINE) && !(PIN_I_SINE))
                {
                    Config.sts = STS_SIN_CMD;
                    break;
                }

                //High voltage
                if (Config.inv.vmv > S_VL_INV_MAX)
                {
                    Config.sts = STS_BUS_OVLO;
                    break;
                }

                break;
            }

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * SAFETY SHUT-DOWN:  shut inverter down if running and error code is
             * set by checks.  Stopping the inverter, by function, also stops 
             * sine.
             * 
             * However, the relay is left on and if its status is set it is also 
             * turned off for safety.  This isn't done in the run-time check
             * because if turned off while the inverter is operating it will
             * cause the start circuit's fuse to open and possibly damage it.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (Config.sts != STS_RUN)
            {
                InverterOFF();
                if (Config.sts == STS_RLY_CMD) setRelay(FALSE);
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * AUTOSTART (BOOT):  check for autostart.  If TRUE the DC bus relay is
         * switched on and then the inverter is started.  A delay is added to 
         * allow the sine section to boot and quiesce.  A failure in any of 
         * these steps will abort the autostart; it's expected that the function
         * where start failed has set Config.sts to the appropriate error code.
         * 
         * If autostart is disabled this code block still executes to set STOP
         * status code.
         * 
         * Last step of boot is to enable the watchdog timer.  This isn't done
         * earlier because of delays inserted in boot process & autostart.
         * Rationale:  any failure prior to this point is not a safety or
         * damage concern.
         * 
         * Note:  it's expected that watchdog config is completed by function
         * Initialize.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts == STS_BOOT)
        {
            SWDTEN = FALSE;                     //Make sure disabled

            //Fetch Vref as starting point for digital control loop
            ReadVoltage(ADC_CH_DAC_5, S_RT_REF, &Config.ref);

            if (ramData[FLASH_AR_AUTO] != S_BL_AUTOD)
            {
                if (setRelay(TRUE))
                {
                    __delay_ms(S_TM_RELAY);
                    if (InverterON())
                    {
                        if (ramData[FLASH_AR_AUTO] == S_BL_AUTOS)
                        {
                            __delay_ms(S_TM_START);
                            if (setSine(TRUE)) Config.sts = STS_RUN;
                        }
                        else Config.sts = STS_RUN;
                    }
                }
            }
            else Config.sts = STS_STOP;

            SWDTEN = TRUE;                      //Enable watchdog timer
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Record last error status to NVR for service.  Don't record under-volt
         * lockout (UVLO) since this will happen on normal shutdown with high
         * probability of NVM corruption.
         * 
         * 19 May 18:  added VCC minimum voltage check as hack to prevent NVM
         * write failures during normal power-downs.  It appears that after UVLO
         * is detected and the inverter turned off there is sufficient energy in
         * both the bridge & aux bulk supply caps to increase voltages and allow
         * a reboot of the MCU that passes voltage checks and attempts restart.
         * Ultimately this needs to be fixed with a 'power good' pin that holds
         * MCU in reset.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((Config.sts >= STS_CODE_ERR) && (ramData[FLASH_AR_STS] != Config.sts))
        {
            ReadVoltage(PIN_I_IVVCCV, S_RT_VCC, &Config.vcc);
            if ((Config.sts != STS_BUS_UVLO) && (Config.vcc.vmv > S_LM_VCC_MIN))
            {
                ramData[FLASH_AR_STS] = Config.sts;
                writeFlashRow((uint16_t) nvData, ramData);
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * LED DISPLAY UDPATE
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        setLED(Config.sts, Config.pwr);


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * FAULT RECOVERY CHECKS:  attempt to resume operation for select fault
         * conditions when they clear.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts > STS_CODE_OK)
        {
            if ((Config.sts == STS_BUS_UVLO) && (Config.inv.vmv > S_VL_INV_RSTART))
                Config.sts = STS_BOOT;
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * CONSOLE PARAMETER MONITORING.  Works like Unix 'top' utility to
         * display system data on console at ~ 1S intervals:
         *  + Inverter DC+
         *  + Vcc sense
         *  + Bridge temp (NTC volts)
         *  + Duty cycle
         *  + S&H current sense
         * 
         * Bit-flag statMonitor controls when monitoring is on and is set with 
         * console command.  For code space efficiency a common format is string
         * is used with all parameters except the last one which includes the
         * line feed (instead of tab) to avoid the overhead of a separate call.
         * The single-byte monTimer, maintained by the system timing ISR, times
         * the update interval which is set by constant S_TM_MONINTVL.
         * 
         * Care should be taken with the use of this feature, particularly when 
         * the digital control loop is active, due to the overhead it imposes.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (statMonitor)
        {
            if (!(monTimer))
            {
                printf(fmtMonvd, lblInv, Config.inv.volt, Config.inv.mv);
                printf(fmtMonvd, lblTemp, Config.tmp.volt, Config.tmp.mv);
                printf(fmtMonvd, lblDc, (unsigned int) (Config.cogt >> 4),\
                        (unsigned int) (Config.cogt & 0x0F));
                printf(fmtMonvd, lblAmp, Config.amps, 0);
                printf(fmtMonlt, lblPwr, Config.pwr, 0);
                monTimer = S_TM_MONINTVL;           //Reset timer
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * CONSOLE SERVICE:  process user input.
         * 
         * DOCUMENT prnResult!!!!!!!!!!!
         * 
         * Important:  do not place any code expected to execute on each loop
         * iteration beyond this point.  The console service handler should 
         * always be placed at bottom of run loop.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (usartStatRx() != USARTSTSRXRDY) continue;
        CmdBuffer = usartGetRx();
        prnResult = TRUE;

        if (!(strcmp(CmdBuffer, CmdList[CMDDSPHLP].str)))
        {
            printf("\n%s v%u.%u, build %u\n%s\n\nCommands:\n", msgName,\
                    VERMAJOR, VERMINOR, VERBUILD, msgCW);
            for (cmdCount = 0; (cmdCount < NUMOFCMDS); ++cmdCount)\
                printf("    %s=%s\n", CmdList[cmdCount].str, CmdList[cmdCount].desc);
        }
        else if (!(stricmp(CmdBuffer, CmdList[CMDDSPSTS].str))) DisplayStatus();
        else if (!(stricmp(CmdBuffer, CmdList[CMDMONOFF].str)))
        {
            clearCmdResult();
            statMonitor = FALSE;
            printCmdResult();
            usartSetEcho(TRUE);
        }
        else if (!(stricmp(CmdBuffer, CmdList[CMDMONON].str)))
        {
            clearCmdResult();
            statMonitor = TRUE;
            monTimer = S_TM_MONINTVL;
            printCmdResult();
            usartSetEcho(FALSE);
        }
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETLPM].str, CMDSIZSTR)))\
            setLoop(CmdBuffer);
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETDCY].str, CMDSIZSTR)))\
            setDutyCycle(CmdBuffer);
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETBLK].str, CMDSIZSTR)))\
            setBlanking(CmdBuffer);
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETAUT].str, CMDSIZSTR)))\
            setAutostart(CmdBuffer);
        else if (!(stricmp(CmdBuffer, CmdList[CMDINVON].str))) InverterON();
        else if (!(stricmp(CmdBuffer, CmdList[CMDINVOFF].str))) InverterOFF();
        else if (!(stricmp(CmdBuffer, CmdList[CMDSINON].str))) setSine(TRUE);
        else if (!(stricmp(CmdBuffer, CmdList[CMDSINOFF].str))) setSine(FALSE);
        else if (!(stricmp(CmdBuffer, CmdList[CMDRLYON].str))) setRelay(TRUE);
        else if (!(stricmp(CmdBuffer, CmdList[CMDRLYOFF].str))) setRelay(FALSE);
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETCUR].str, CMDSIZSTR)))\
            setCurrentLimit(CmdBuffer);
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETVRF].str, CMDSIZSTR)))\
            setVref(CmdBuffer);
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETPRG].str, CMDSIZSTR)))\
            setPrg(CmdBuffer);
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETTYP].str, CMDSIZSTR)))\
            setLoopType(CmdBuffer);
        else if (!(stricmp(CmdBuffer, CmdList[CMDRSTPIC].str)))
        {
            //Reset system; first do clean stop if running
            if (!(COG1ASD0bits.ASE)) InverterOFF();
            SWDTEN = FALSE;             //Disable so watchdog doesn't erroneously reboot
            PCON = S_LM_PCON_BT;        //Clear PCON bits to clear error trap on boot
            __delay_ms(S_TM_RST);
            RESET();
        }
        else if (!(stricmp(CmdBuffer, CmdList[CMDCLRERR].str)))
        {
            clearCmdResult();
            ramData[FLASH_AR_STS] = STS_UND;
            writeFlashRow((uint16_t) nvData, ramData);
            printCmdResult();
        }
        else printf(msgHuh);

        usartSetRx(ENABLE);             //Look for new user input
    }
} //End function main


/* Function DisplayStatus.  Prints status & configuration:
 * 
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void DisplayStatus()
{
    ReadVoltage(ADC_CH_DAC_4, S_RT_VDD, &Config.vdd);
    ReadVoltage(ADC_CH_DAC_5, S_RT_REF, &Config.ref);
    ReadVoltage(PIN_I_IVVCCV, S_RT_VCC, &Config.vcc);

    printf("\n%s%s, NVM=%s, Mon=%s", lblSys, StsList[Config.sts].desc,\
            StsList[ramData[FLASH_AR_STS]].desc, ((statMonitor) ? msgOn : msgOff));
    printf("\nMCU: STKOVF=%u, STKUNF=%u, RWDT=%u, RMCLR=%u, RI=%u, POR=%u, BOR=%u, TO=%u, PD=%u",\
            (unsigned int) STKOVF, (unsigned int) STKUNF, (unsigned int) nRWDT,\
            (unsigned int) nRMCLR, (unsigned int) nRI, (unsigned int) nPOR,\
            (unsigned int) nBOR, (unsigned int) nTO, (unsigned int) nPD);
    printf("\nControl/Ack: relay=%s/%s, inverter=%s, sine=%s/%s, auto=",\
            ((PIN_O_RELAY) ? msgOn : msgOff), ((PIN_I_RELAY) ? msgOn : msgOff),\
            ((COG1ASD0bits.ASE) ? msgOff : msgOn), ((PIN_O_SINE) ? msgOn : msgOff),\
            ((PIN_I_SINE) ? msgOn : msgOff));
    if (ramData[FLASH_AR_AUTO] == S_BL_AUTOI) printf(msgInv);
    else if (ramData[FLASH_AR_AUTO] == S_BL_AUTOS) printf(msgSine);
    else printf(msgOff);
    printf("\nLoop: %c, Type=%c, DC A/M=%u%%/M=%u.%uuS, Vref V/R=%u.%3.3u/%d, %s%u.%3.3u, PRG=%d",\
            ((statOpen) ? S_CH_OPEN : S_CH_CLOSED), ((statDigital) ? S_CH_DIGITAL : S_CH_ANALOG),\
            Config.odcp, (unsigned int) (Config.cogt >> 4), (unsigned int) (Config.cogt & 0x0F),\
            Config.ref.volt, Config.ref.mv, (int) ((DAC5REFH << 8) | (DAC5REFL)),\
            lblVccs, Config.sen.volt, Config.sen.mv, (int) PRG3CON2);
    printf("\nCurrent: limit=%d, %s%u, %s%u.%3.3u, %s%d, %s%d",\
            (int) ((DAC1REFH << 8) | (DAC1REFL)), lblBtime, Config.btim, lblShcs, 
            Config.ics.volt, Config.ics.mv, lblAmp, Config.amps, lblPwr, Config.pwr);
    printf("\nVoltages: %s%u.%3.3u, Vcc=%u.%3.3u, Vdd=%u.%3.3u, %s%u.%3.3u",\
            lblInv, Config.inv.volt, Config.inv.mv, Config.vcc.volt, Config.vcc.mv,\
            Config.vdd.volt, Config.vdd.mv, lblTemp, Config.tmp.volt, Config.tmp.mv);

    printf(msgNewl);

    return;
} //End function DisplayStatus


/* Function setLoop.  Sets control loop open or closed.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setLoop(char *setstr)
{
    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    if (*setstr < 97) *setstr += CMDLOWCAS;             //Change to lower case for eval
    if (*setstr == CMDCLOSLP)                           //Closed
    {
        if (statOpen)                                   //Do nothing if already closed loop
        {
            if (!(statDigital))                         //Use analog loop
            {
                COG1FIS1 = CFG_COG_1_FEI1_CLOOP;        //Connect CLC, ok if DC decreases
                if (!(COG1ASD0bits.ASE)) chgDutyCycle(Config.oldc, S_DC_MAX);   //Set max allowable DC
            }
            statOpen = FALSE;
        }
    }
    else if (*setstr == CMDOPENLP)                      //Open
    {
        if (!(statOpen))                                //Do nothing if already open loop
        {
            if (!(COG1ASD0bits.ASE))
            {
                getCOGtime();                           //Get DC while still closed
                PWM5DC = Config.cogc;                   //Load this as starting value in PWM
                PWM5LDCONbits.LDA = TRUE;
                COG1FIS1 = CFG_COG_1_FEI1_OLOOP;        //Disconnect CLC, but no DC change
                chgDutyCycle(Config.cogc, Config.oldc); //Move to OL DC
            }
            else COG1FIS1 = CFG_COG_1_FEI1_OLOOP;       //Even if off must disconnect CLC
            statOpen = TRUE;
        }
    }
    else setCmdErr();

    if (ramData[FLASH_AR_OPENL] != (unsigned char) statOpen)
    {
        ramData[FLASH_AR_OPENL] =(unsigned char) statOpen;
        writeFlashRow((uint16_t) nvData, ramData);
    }
    printCmdResult();

    return;
} //End function setLoop


/* Function setSine.  Sets output pin to enable / disable digital sine AC output.
 * 
 * Returns:  boolean TRUE if ack received, FALSE if not.  If FALSE then output
 *           pin is set false regardless of cmd and Config.sts contains error
 *           code.
 *
 * Parameters:  unsigned char cmd, TRUE=on, FALSE=off
 *
*/
bit setSine(unsigned char cmd)
{
    clearCmdResult();

    if (cmd)                                //Enable
    {
        PIN_O_SINE = TRUE;
        //Wait S_LM_SIN_ON x S_TM_ISN_ON increments for ack.
        for (cmd = S_LM_SIN_ON; (!(PIN_I_SINE) && (cmd)); --cmd, __delay_us(S_TM_SIN_ON));
    }
    else                                    //Disable
    {
        PIN_O_SINE = FALSE;
        //Wait S_LM_SIN_OFF x S_TM_SIN_OFF increments for ack.
        for (cmd = S_LM_SIN_OFF; ((PIN_I_SINE) && (cmd)); --cmd, __delay_us(S_TM_SIN_OFF));
    }

    //If cmd==0 assume timeout occurred; set output off & error return.
    if (!(cmd))
    {
        PIN_O_SINE = FALSE;
        Config.sts = STS_SIN_CMD;
        setCmdErr();
    }

    printCmdResult();

    return(cmdGood);
} //End function setSine


/* Function setRelay.  Sets output pin to enable / disable for SSR control.
 * 
 * Returns:  boolean TRUE if ack received, FALSE if not.  If FALSE then output
 *           pin is set false regardless of cmd and Config.sts contains error
 *           code.
 *
 * Parameters:  unsigned char cmd, TRUE=on, FALSE=off
 *
*/
bit setRelay(unsigned char cmd)
{
    clearCmdResult();

    if (cmd)                                //Enable
    {
        PIN_O_RELAY = TRUE;
        //Wait S_LM_RLY_ON x S_TM_RLY_ON increments for ack.
        for (cmd = S_LM_RLY_ON; (!(PIN_I_RELAY) && (cmd)); --cmd, __delay_us(S_TM_RLY_ON));
    }
    else                                    //Disable
    {
        PIN_O_RELAY = FALSE;
        //Wait S_LM_RLY_OFF x S_TM_RLY_OFF increments for ack.
        for (cmd = S_LM_RLY_OFF; ((PIN_I_RELAY) && (cmd)); --cmd, __delay_us(S_TM_RLY_OFF));
    }

    //If cmd==0 assume timeout occurred; set output off & error return.
    if (!(cmd))
    {
        PIN_O_RELAY = FALSE;
        Config.sts = STS_RLY_CMD;
        setCmdErr();
    }

    printCmdResult();

    return(cmdGood);
} //End function setRelay


/* Function setAutostart.  Sets autostart and updates NV memory.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setAutostart(char *setstr)
{
    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    if (*setstr < 97) *setstr += CMDLOWCAS;             //Change to lower case for eval
    if (*setstr == CMDINVERT)                           //Inverter only
    {
        if (ramData[FLASH_AR_AUTO] != S_BL_AUTOI)
        {
            ramData[FLASH_AR_AUTO] = S_BL_AUTOI;
            writeFlashRow((uint16_t) nvData, ramData);
        }
    }
    else if (*setstr == CMDINSINE)                      //Inverter & sine
    {
        if (ramData[FLASH_AR_AUTO] != S_BL_AUTOS)
        {
            ramData[FLASH_AR_AUTO] = S_BL_AUTOS;
            writeFlashRow((uint16_t) nvData, ramData);
        }
    }
    else if (*setstr == CMDDSABLE)                      //Disable
    {
        if (ramData[FLASH_AR_AUTO] != S_BL_AUTOD)
        {
            ramData[FLASH_AR_AUTO] = S_BL_AUTOD;
            writeFlashRow((uint16_t) nvData, ramData);
        }
    }
    else setCmdErr();

    printCmdResult();

    return;
} //End function setAutostart


/* Function setLoopType.  Sets control loop type to analog or digital.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setLoopType(char *setstr)
{
    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    if (*setstr < 97) *setstr += CMDLOWCAS;             //Change to lower case for eval
    if (*setstr == CMDLOOPAN)                           //Analog
    {
        if (ramData[FLASH_AR_LOOP_D] != FALSE)
        {
            ramData[FLASH_AR_LOOP_D] = FALSE;
            writeFlashRow((uint16_t) nvData, ramData);
        }
    }
    else if (*setstr == CMDLOOPDG)                      //Digital
    {
        if (ramData[FLASH_AR_LOOP_D] != TRUE)
        {
            ramData[FLASH_AR_LOOP_D] = TRUE;
            writeFlashRow((uint16_t) nvData, ramData);
        }
    }
    else setCmdErr();

    statDigital = ramData[FLASH_AR_LOOP_D];

    printCmdResult();

    return;
} //End function setLoopType


/* Function setDutyCycle.  Sets the open loop duty cycle.  If in closed loop
 * the entered value is bounds checked and stored in object Config for use 
 * when open loop is set.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setDutyCycle(char *setstr)
{
    unsigned int dcp;                       //User entered duty cycle as %
    unsigned int oldcSave;                  //Value of oldc prior to change

    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    dcp = atoi(setstr);
    if ((dcp < S_DC_MIN_P) || (dcp > S_DC_MAX_P)) setCmdErr();      //Bounds check
    else
    {
        oldcSave = Config.oldc;
        Config.odcp = dcp;
        Config.oldc = (((dcp * CFG_PWM_5_FRQ) / 100) + 1);
        if (statOpen & !(COG1ASD0bits.ASE)) chgDutyCycle(oldcSave, Config.oldc);

        //Update NVM
        if (Config.odcp != (unsigned int) ramData[FLASH_AR_OLDC_P])
        {
            ramData[FLASH_AR_OLDC_P] = (unsigned char) Config.odcp;
            ramData[FLASH_AR_OLDC] = (unsigned char) Config.oldc;
            writeFlashRow((uint16_t) nvData, ramData);
        }
    }

    printCmdResult();

    return;
} //End function setDutyCycle


/* Function setBlanking.  Sets the blanking time on rising edge of COG to
 * prevent spurious current sense falling events.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setBlanking(char *setstr)
{
    unsigned int dc;                       //User entered duty cycle

    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    if ((dc = atoi(setstr)) > CFG_PWM_11_PERIOD) setCmdErr();       //Bounds check
    else
    {
        Config.btim = dc;
        PWM11DC = dc;
        PWM11LDCONbits.LDA = TRUE;
    }

    printCmdResult();

    return;
} //End function setBlanking


/* Function setPrg.  Sets the PRG ramp rate (dv/dt).  Used for testing control
 * loop gain.  Dynamically set regardless of loop state and does not survive
 * a reboot.  The entered value is bounds checked.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setPrg(char *setstr)
{
    int rate;                               //Integer representation of entered value.

    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    rate = atoi(setstr);
    if ((rate < S_LM_PRGLOW) || (rate > S_LM_PRGHIG)) setCmdErr();  //Bounds check
    else PRG3CON2 = (unsigned char) rate;

    printCmdResult();

    return;
} //End function setPrg


/* Function setCurrentLimit.  Sets the current limit by updating the DAC with 
 * user-supplied value.  The entered value is bounds checked.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setCurrentLimit(char *setstr)
{
    unsigned int limit;                     //New current limit value

    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    limit = atoi(setstr);
    if ((limit < S_LM_CLLOW) || (limit > S_LM_CLHIG)) setCmdErr();  //Bounds check
    else
    {
        DAC1REFH = (unsigned char) (limit >> 8);
        DAC1REFL = (unsigned char) (limit & 0x00FF);
        DAC1LD = TRUE;
        while (DAC1LD);
    }

    printCmdResult();

    return;
} //End function setCurrentLimit


/* Function setVref.  Sets the Vref value used in the Vcc voltage regulation
 * control loop.  The DAC is updated with the user entered value and stored
 * in NVM.  The entered value is bounds checked.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setVref(char *setstr)
{
    unsigned int vref;                      //New value

    clearCmdResult();

    for (setstr += CMDSIZSTR; (*setstr == CMDWSPACE); ++setstr);    //Get thru white space
    vref = atoi(setstr);
    if ((vref < S_LM_VRLOW) || (vref > S_LM_VRHIG)) setCmdErr();    //Bounds check
    else
    {
        //Split 16-bit integer into high/low bytes
        ramData[FLASH_AR_VREF_H] = (unsigned char) (vref >> 8);
        ramData[FLASH_AR_VREF_L] = (unsigned char) (vref & 0x00FF);

        //Update DAC
        DAC5REFH = ramData[FLASH_AR_VREF_H];
        DAC5REFL = ramData[FLASH_AR_VREF_L];
        DAC5LD = TRUE;
        while (DAC5LD);

        writeFlashRow((uint16_t) nvData, ramData);                  //Update NVM
    }

    //Always update Vref value so current for digital control loop
    ReadVoltage(ADC_CH_DAC_5, S_RT_REF, &Config.ref);

    printCmdResult();

    return;
} //End function setVref


/* Function InverterON.  Starts inverter.
 * 
 * Returns:  boolean TRUE if started, FALSE if not; Config.sts contains error code.
 *
 * Parameters:  none
 *
*/
bit InverterON()
{
    unsigned int finalDC;
    unsigned char result;

    setCmdErr();

    for (result = 0; (TRUE); )
    {
        if (COG1ASD0bits.ASE && ((Config.sts == STS_STOP) || (Config.sts == STS_BOOT)))
        {
            if (Config.inv.vmv < S_VL_INV_MIN)
            {
                result = STS_BUS_UVLO;
                break;
            }

            if (result = chargeBScaps()) break;

            if (statOpen) finalDC = Config.oldc; else finalDC = S_DC_MAX;
            PWM5DC = S_DC_MIN;                      //Soft start with min DC
            PWM5LDCONbits.LDA = TRUE;
            COG1ASD0bits.ASE = FALSE;               //Output on
            chgDutyCycle(S_DC_MIN, finalDC);        //Increase to max DC

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Delay allows duty cycle to catch-up since during start the HV
             * secondary impedance will be very low until bulk caps are charged.
             * Because of this the COG DC will be limited by the current sense
             * loop.  If DC is not as expected after delay then something serious
             * is likely wrong and inverter is stopped.
             * 
             * Note that this method is effective for open loop / DC transformer
             * operation but not for closed loop and requires further development
             * if that mode is to be used.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            __delay_ms(S_TM_INVON_COG);
            getCOGtime();

            if (statOpen)
            {
                if (Config.cogc < Config.oldc) result = STS_COG_TIME;
            }
            else if (Config.cogc == 0) result = STS_COG_TIME;

            if (!(result))                          //Start successful
            {
                result = STS_RUN;
                setCmdOK();
            }
            else COG1ASD0bits.ASE = TRUE;           //Start failed, inverter off
        }
        break;
    }

    if (result) Config.sts = result;
    printCmdSys();

    return(cmdGood);
} //End function InverterON


/* Function InverterOFF.  Stops inverter.
 * 
 * Returns:  none.  If Config.sts is set to STOP provided it doesn't contain an 
 *           error code.
 *
 * Parameters:  none
 *
*/
void InverterOFF()
{
    COG1ASD0bits.ASE = TRUE;            //Inverter off
    if (Config.sts < STS_CODE_ERR) Config.sts = STS_STOP;   //Don't clear error

    PIN_O_SINE = FALSE;                 //Then disable sine bridge

    printCmdSys();

    return;
} //End function InverterOFF


/* Function ReadVoltage.  Reads the value from the analog pin channel, interprets 
 * as a scaled voltage, and stores in the volts object pointed to by rv.
 * 
 * The ADC must be pre-configured to use the FVR x 4 voltage reference (4.096V).
 * 
 * No error checking is performed and only a single sample is taken.  The voltage
 * is stored in 'composite' form for easy integer evaluation of the complete
 * value (e.g. V/mV) by storing the integer value in the 1,000 & 10,000 place
 * of a 16-bit integer and the mV in the 100, 10, and 1.  The discrete volts
 * and mV are also stored in separate integers.
 * 
 * Returns:  none.  The following values are populated in object rv:
 *      vmv=composite value
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=unmodified value read from ADC
 *
 * Parameters:  channel=analog channel.  The pin must be properly configured.
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.
 *
*/
void ReadVoltage(unsigned char channel, unsigned char scale, volts *rv)
{
    static union {
        unsigned char result[SIZ_INT_ADC];
        unsigned int value;
    } adc;

    ADCON0 = (unsigned) (channel | S_EN_ADC);   //Connect to channel & enable ADC
    __delay_us(FRV_ADC_AQT);                    //Refresh sample & hold cap
    ADCON0bits.GO = TRUE;                       //Start conversion
    while (ADCON0bits.GO);                      //Wait for completion
    adc.result[POS_HIG_ADC] = ADRESH;
    adc.result[POS_LOW_ADC] = ADRESL;

    rv->adc = adc.value;                        //Save raw ADC value to user object
    adc.value *= scale;                         //Scale value
    rv->vmv = (((adc.value / FRV_ADC_V1V) * 1000) + ((adc.value % FRV_ADC_V1V) * 4));
    rv->volt = rv->vmv / 1000;
    rv->mv = rv->vmv % 1000;

    return;
} //End function ReadVoltage


/* Function CalcAvg.  Calculates the arithmetic mean of a sample set of 
 * previously acquired values.  No bounds checking is performed.
 * 
 * The value to be averaged is expected to represent the sum of raw ADC
 * conversions; they should not be scaled.  The average is calculated and then
 * scaled by CalcAvg.  The results are stored in a user supplied object of type
 * volts.
 * 
 * Returns:  none.  The following values are populated in object rv.  All 
 * represent averaged values:
 *      vmv=composite value
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=value read from ADC
 *
 * Parameters:  sum=raw summed ADC values
 *              count=number of summed values
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.
 *
*/
void CalcAvg(unsigned int sum, unsigned char count, unsigned char scale, volts *rv)
{
    rv->adc = ((sum / count) * scale);
    rv->vmv = (((rv->adc / FRV_ADC_V1V) * 1000) + ((rv->adc % FRV_ADC_V1V) * 4));
    rv->volt = rv->vmv / 1000;
    rv->mv = rv->vmv % 1000;

    return;
} //End function CalcAvg


/* Function getCOGtime.  Calculates the on time of the COG and stores in module
 * object Config (type invcf).  COG on-time capture is performed using timer5 with
 * gate control and single pulse modes.  The base timer configuration is 
 * expected to be completed at boot.
 * 
 * The calculated on-time is stored in a single byte (cogt).  The integer component of 
 * the value is calculated by dividing the timer low byte count by the count for 
 * 1uS.  The modulus of the timer value is then multiplied by 100 and divided by 
 * the maximum count (10uS).
 * 
 * Returns:  none.  The following values are populated in Config:
 *      cogc=value from timer, each increment=62.5nS
 *      cogt=upper nibble uS, lower nibble=100's nS
 * 
 * cogc and cogt are set to zero if the timer's lower register exceeds FGCT_MAXCNT.
 *
 * Parameters:  none
 *
*/
void getCOGtime()
{
    //Clear timer: high byte is not used
    TMR5 = 0;

    T5GGO_nDONE = TRUE;                 //Start capture
    __delay_us(FGCT_WAITTM);            //Fixed wait time for completion
    if (T5GGO_nDONE)                    //Problem or COG is off
    {
        T5GGO_nDONE = FALSE;            //Stop timing
        TMR5L = (FGCT_MAXCNT + 1);      //Set high value & let remaining logic clear
    }

    if (TMR5L > FGCT_MAXCNT)            //Invalid DC - clear values
    {
        Config.cogc = 0;
        Config.cogt = 0;
    }
    else                                //Calculate valid DC
    {
        Config.cogc = TMR5;
        Config.cogt = ((TMR5L / FGCT_1USCNT) << 4);
        Config.cogt |= (((TMR5L % FGCT_1USCNT) * 100) / FGCT_MAXCNT);
    }

    return;
} //End function getCOGtime


/* Function chgDutyCycle.  Moves the duty cycle of the bridge from the start
 * to final value incrementally to avoid abrupt output changes.
 * 
 * Returns:  start=16-bit starting DC value
 *           final=16-bit final DC value
 *
 * Parameters:  none
 *
*/
void chgDutyCycle(unsigned int start, unsigned int final)
{
    while (TRUE)
    {
        PWM5DC = start;
        PWM5LDCONbits.LDA = TRUE;
        if (start == final) break;

        if (start < final) ++start; else --start;
        __delay_us(S_TM_DCDELAY);
    }

    return;
} //End function chgDutyCycle


/* Function chargeBScaps.  Charges bootstrap capacitors by alternately turning on
 * the low-side switch of each bridge channel.  The COG is disconnected from
 * low side pin to facilitate and during the process short circuit monitoring
 * is performed.
 * 
 * This function does not perform state checks and expects to be called from a
 * higher level function that ensures the proper operating state.
 * 
 * Short circuit detection is accomplished by directly reading the output of each
 * current sense comparator.  This is done because the over-current event ISR
 * detects a state change only; not a steady state.  Short detection is performed
 * after the charge cycle is expected to complete since the bootstrap capacitors
 * can draw several amperes of current depending on size & current limiting.
 * 
 * Returns:  0 if successful or non-zero value representing error.
 *
 * Parameters:  none
 *
*/
unsigned char chargeBScaps()
{
    unsigned char result;

    for (result = 0; (TRUE); )
    {
        //Charge A-side
        PIN_O_SIDE_A_L = FALSE;
        PPS_P_SIDE_A_L = PPS_M_LAT_RC5;         //Take control of low-side channel pin
        PIN_O_SIDE_A_L = TRUE;                  //Turn switch on
        //Caps reach full charge < 1uS
        NOP();
        NOP();
        NOP();
        NOP();
#ifndef _NO_LIMITS
        if (!(S_MP_CSENS)) result = STS_CHL_BSAF;
#endif
        PIN_O_SIDE_A_L = FALSE;                 //Turn switch off
        PPS_P_SIDE_A_L = PPS_M_COG1_D;          //Return pin control to COG
        if (result) break;

        //Charge B-side
        PIN_O_SIDE_B_L = FALSE;
        PPS_P_SIDE_B_L = PPS_M_LAT_RB5;         //Take control of low-side channel pin
        PIN_O_SIDE_B_L = TRUE;                  //Turn switch on
        //Caps reach full charge < 1uS
        NOP();
        NOP();
        NOP();
        NOP();
#ifndef _NO_LIMITS
        if (!(S_MP_CSENS)) result = STS_CHL_BSBF;
#endif
        PIN_O_SIDE_B_L = FALSE;                 //Turn switch off
        PPS_P_SIDE_B_L = PPS_M_COG1_C;          //Return pin control to COG

        break;
    }

    return(result);
} //End function chargeBScaps


/* Function writeFlash.  Writes a row of flash.
 * 
 * Returns:  none
 *
 * Parameters:  data=pointer to array of data to write to flash
 *              addr=14-bit address of row in flash to write.  Must point to
 *                  start of array.
 *
*/
void writeFlashRow(uint16_t addr, unsigned char *data)
{
    unsigned char count;

    eraseFlashRow(addr);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Setup to write data to flash.  Interrupts must be disabled for entire
     * sequence.  Set LWLO to load data to write latches and then to flash on
     * last byte.
     * 
     * Load high address register here & only once.  Update low register for 
     * individual bytes to write within row.
     * 
     * Variable count represents array element to write.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    di();                               //Interrupts disabled for entire sequence.
    WREN = TRUE;                        //Enable write
    CFGS = FALSE;                       //Program memory
    FREE = FALSE;                       //Write data
    LWLO = TRUE;                        //Load data latches
    PMADRH = (unsigned char) (addr >> 8);

    for (count=0; (count < FLASHROWSZ); ++count)
    {
        PMADRL = (unsigned char) (addr + count);
        PMDATL = *(data + count);
        unlockSeq();                    //Unlock for write
    }

    LWLO = FALSE;                                   //Write row
    unlockSeq();                                    //Write flash

    //All done:  disable writes & re-enable interrupts
    WREN = FALSE;
    ei();

    return;
} //End function writeFlashRow


/* Function eraseFlashRow.  Erases a row of program flash memory.  Interrupts
 * are disabled during the erase.
 * 
 * Returns:  none
 *
 * Parameters:  addr=14-bit pointer to starting address for row.
 *
*/
void eraseFlashRow(uint16_t addr)
{
    PMADRL = (unsigned char) addr;
    PMADRH = (unsigned char) (addr >> 8);

    //Erase row
    di();                               //Disable interrupts for duration of function
    CFGS = FALSE;                       //Program memory
    FREE = TRUE;                        //Erase row
    WREN = TRUE;                        //Enable erase
    unlockSeq();                        //Unlock for erase
    WREN = FALSE;                       //Disable erase
    ei();                               //Enable interrupts

    return;
} //End function eraseFlashRow


/* Function setLED.  Sets the front panel LEDs according to the status & power
 * level.
 * 
 * Returns:  none
 *
 * Parameters:  status=unsigned char representing status code
 *              power = unsigned int representing inverter power
 *
*/
void setLED(unsigned char status, unsigned int power)
{
    unsigned char ledmap;                   //Local storage for StsList[].ledmap

    //Start with all pins off
    PIN_O_FAULT = FALSE;
    PIN_O_RUN = FALSE;
    PIN_O_BOOST = FALSE;
    PIN_O_L100 = FALSE;
    PIN_O_L75 = FALSE;
    PIN_O_L50 = FALSE;
    PIN_O_L25 = FALSE;

    if (status == STS_RUN)
    {
        //Set run LED
        PIN_O_FAULT = FALSE;
        PIN_O_RUN = TRUE;

        //Set 100% LED to 100% or boost
        if (Config.pwr > S_LM_PWR_BOOST) PIN_O_BOOST = TRUE;
        else if (Config.pwr > S_LM_PWR_100) PIN_O_L100 = TRUE;

        //Remaining power LEDs function as bar graph
        if (Config.pwr > S_LM_PWR_75) PIN_O_L75 = TRUE;
        if (Config.pwr > S_LM_PWR_50) PIN_O_L50 = TRUE;
        if (Config.pwr > S_LM_PWR_25) PIN_O_L25 = TRUE;
    }
    else
    {
        //Fetch led map for from status to reduce repetitive access times
        ledmap = StsList[status].ledmap;

        //Set pins according to status map to display fault & code
        PIN_O_FAULT = (ledmap & S_MK_FAULT);
        PIN_O_RUN = ((ledmap & S_MK_RUN) >> 1);
        PIN_O_BOOST = ((ledmap & S_MK_BOOST) >> 2);
        PIN_O_L100 = ((ledmap & S_MK_L100) >> 3);
        PIN_O_L75 = ((ledmap & S_MK_L75) >> 4);
        PIN_O_L50 = ((ledmap & S_MK_L50) >> 5);
        PIN_O_L25 = ((ledmap & S_MK_L25) >> 6);
    }

    return;
} //End function setLED


/*Function InterruptExec.  MPU interrupt handler.  Manages all enabled
 * interrupts:
 *      + USART Rx
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void interrupt InterruptExec(void)
{
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Caretaker.  Not sure if this will be needed but leave for future.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (TMR2IE & TMR2IF)
    {
        if (monTimer) --monTimer;
        TMR2IF = FALSE;
    }

    usartISR();

    return;
} //End function InterruptExec


/*Function Initialize.  Configures pins & peripherals on the MPU.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * System clock & watchdog config
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OSCCON = CFG_OSC_C;                 //Configure system clock
    while (!(HFIOFL));                  //Wait for oscillator to stabilize
    WDTCON = CFG_WDT_CON;               //Watchdog


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Retrieve NVM settings & assign.  Do before MCU & NVM checks.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    memcpy((void *) ramData, (void *) nvData, FLASHROWSZ);
    //Set loop status
    if (ramData[FLASH_AR_OPENL]) statOpen = TRUE; else statOpen = FALSE;
    //Set digital control loop status
    if (ramData[FLASH_AR_LOOP_D]) statDigital = TRUE; else statDigital = FALSE;
    Config.odcp = (unsigned int) ramData[FLASH_AR_OLDC_P];  //Starting OL DC (%) is run value
    Config.oldc = (unsigned int) ramData[FLASH_AR_OLDC];    //Starting OL DC (16-bit) is run value


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * If booting for anything other than power-up or reset command set status
     * to prevent start and preserve evidence.
     * 
     * Also check for corrupt NVM data:  validate loop duty cycle as proxy.
     * Future improvement would be to checksum once design is finished and 
     * checksum overhead can be factored.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (((PCON & S_LM_PCON_EX) != S_LM_PCON_BT) || \
            ((STATUS & S_LM_STAT_EX) != S_LM_STAT_BT)) Config.sts = STS_CPU;
    else if (Config.oldc > S_DC_MAX) Config.sts = STS_NVM_INV;
    else Config.sts = STS_BOOT;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Important but miscellaneous settings that must be set at boot and aren't
     * part of a larger group.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    statMonitor = FALSE;                //Set monitoring off
    prnResult = FALSE;                  //Don't print command results
    OPTION_REGbits.nWPUEN = FALSE;      //Individual pin control over weak pull-ups
    Config.btim = CFG_PWM_11_DUTYC;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set port configuration.  Ensure all output latches are clear to prevent
     * startup glitches.
     * 
     * Set LED status for boot.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PORTA = 0;
    PORTB = 0;
    PORTC = 0;
    //Port A
    TRISA = IO_PORTA;
    ANSELA = AD_PORTA;
    WPUA = WP_PORTA;
    ODCONA = OD_PORTA;
    SLRCONA = SR_PORTA;
    INLVLA = IL_PORTA;
    //Port B
    TRISB = IO_PORTB;
    ANSELB = AD_PORTB;
    WPUB = WP_PORTB;
    ODCONB = OD_PORTB;
    SLRCONB = SR_PORTB;
    INLVLB = IL_PORTB;
    //Port C
    TRISC = IO_PORTC;
    ANSELC = AD_PORTC;
    WPUC = WP_PORTC;
    ODCONC = OD_PORTC;
    SLRCONC = SR_PORTC;
    INLVLC = IL_PORTC;

    setLED(Config.sts, 0);


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Map peripheral pins
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PPS_P_SIDE_A_H = PPS_M_COG1_A;          //COG1 channel A
    PPS_P_SIDE_A_L = PPS_M_COG1_D;          //COG1 channel D
    PPS_P_SIDE_B_H = PPS_M_COG1_B;          //COG1 channel B
    PPS_P_SIDE_B_L = PPS_M_COG1_C;          //COG1 channel C
    T5GPPS = PPS_M_TMR_5G;                  //Timer5 gate


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Initialize USART for command console
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    usartSetRx(ENABLE);


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure timer2 (caretaker) but don't start
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T2CLKCON = CFG_TMR_2_CLK;
    T2CON = CFG_TMR_2_CON;
    T2HLT = CFG_TMR_2_HLT;
    T2RST = CFG_TMR_2_RST;
    T2PR = CFG_TMR_2_PR;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure timer5 to measure on-time (duty cycle) of COG1.  Function 
     * getCOGtime is dependent on this.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T5CON = CFG_TMR_5_C;
    T5GCON = CFG_TMR_5_G;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure FVR & DACs
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    FVRCON = CFG_FVR_C;
    __delay_us(FI_FVR_DLY);

    DAC1CON0 = CFG_DAC_1_C;
    DAC1REFH = CFG_DAC_1_VH;
    DAC1REFL = CFG_DAC_1_VL;
    DAC1LD = TRUE;                      //Load reference value

    DAC5CON0 = CFG_DAC_5_C;
    DAC5REFH = ramData[FLASH_AR_VREF_H];
    DAC5REFL = ramData[FLASH_AR_VREF_L];
    DAC5LD = TRUE;                      //Load reference value

    DAC7CON0 = CFG_DAC_7_C;
    DAC7REF = CFG_DAC_7_V;

    DAC4CON0 = CFG_DAC_4_C;
    DAC4REF = CFG_DAC_4_V;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure error amp (OPA3)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OPA3CON = CFG_OPA_3_C;
    OPA3NCHS = CFG_OPA_3_N;
    OPA3PCHS = CFG_OPA_3_P;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure over-current sense comparator (CM2)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM2CON0 = CFG_CM_2_C0;
    CM2CON1 = CFG_CM_2_C1;
    CM2NSEL = CFG_CM_2_N;
    CM2PSEL = CFG_CM_2_P;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure ADC baseline operation.  Clock, +/- references, auto-conversion,
     * and justification don't change.  Enable and default connect to DAC4 to
     * calibrate & obtain Vdd values.
     * 
     * Rule:  no assumptions should be made about which analog channel the ADC
     * is connected to.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0 = (ADC_CH_DAC_4 | S_EN_ADC);
    ADCON1 = ADCCFGC1;
    ADCON2 = ADCCFGC2;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM5 as primary PWM signal
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM5INTE = PWM5CIE;
    PWM5CLKCON = PWM5CCL;
    PWM5LDCON = PWM5CLD;
    PWM5OFCON = PWM5COT;
    PWM5PH = PWM5CPH;
    PWM5DC = S_DC_MAX;
    PWM5PR = CFG_PWM_5_FRQ;
    PWM5OF = PWM5COC;
    PWM5TMR = PWM5CTM;
    PWM5CON = PWM5CFG;                      //Start

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM6 as slave to PWM5 to control minimum duty cycle
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM6INTE = PWM6CIE;
    PWM6CLKCON = PWM6CCL;
    PWM6LDCON = PWM6CLD;
    PWM6OFCON = PWM6COT;
    PWM6PH = PWM6CPH;
    PWM6DC = PWM6CDC;
    PWM6PR = PWM6CPD;
    PWM6OF = PWM6COC;
    PWM6TMR = PWM6CTM;
    PWM6CON = PWM6CFG;                      //Start

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM11 as slave to PWM5 to time current sense blanking
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM11INTE = CFG_PWM_11_INTE;
    PWM11CLKCON = CFG_PWM_11_CLK;
    PWM11LDCON = CFG_PWM_11_LDCON;
    PWM11OFCON = CFG_PWM_11_OFCON;
    PWM11PH = CFG_PWM_11_PHASE;
    PWM11DC = CFG_PWM_11_DUTYC;
    PWM11PR = CFG_PWM_11_PERIOD;
    PWM11OF = CFG_PWM_11_OFFSET;
    PWM11TMR = CFG_PWM_11_TIEMR;
    PWM11CON = CFG_PWM_11_CON;              //Start


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CLC1 as OR gate to control COG1 closed loop falling event
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC1SEL0 = CFG_CLC_1_SEL_0;
    CLC1SEL1 = CFG_CLC_1_SEL_1;
    CLC1SEL2 = CFG_CLC_1_SEL_2;
    CLC1SEL3 = CFG_CLC_1_SEL_3;
    CLC1GLS0 = CFG_CLC_1_GLS_0;
    CLC1GLS1 = CFG_CLC_1_GLS_1;
    CLC1GLS2 = CFG_CLC_1_GLS_2;
    CLC1GLS3 = CFG_CLC_1_GLS_3;
    CLC1POL = CFG_CLC_1_POL;
    CLC1CON = CFG_CLC_1_CON;
    LC1EN = TRUE;                           //Enable CLC1

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CLC3 as OR gate to control COG1 current sense falling event
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC3SEL0 = CFG_CLC_3_SEL_0;
    CLC3SEL1 = CFG_CLC_3_SEL_1;
    CLC3SEL2 = CFG_CLC_3_SEL_2;
    CLC3SEL3 = CFG_CLC_3_SEL_3;
    CLC3GLS0 = CFG_CLC_3_GLS_0;
    CLC3GLS1 = CFG_CLC_3_GLS_1;
    CLC3GLS2 = CFG_CLC_3_GLS_2;
    CLC3GLS3 = CFG_CLC_3_GLS_3;
    CLC3POL = CFG_CLC_3_POL;
    CLC3CON = CFG_CLC_3_CON;
    LC3EN = TRUE;                           //Enable CLC3


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure complementary output generator 1 to drive full bridge
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    COG1DBR = CFG_COG_1_RE_DEADB;
    COG1DBF = CFG_COG_1_FE_DEADB;
    COG1BLKR = CFG_COG_1_RE_BLANK;
    COG1BLKF = CFG_COG_1_FE_BLANK;
    COG1PHR = CFG_COG_1_RE_PHASE;
    COG1PHF = CFG_COG_1_FE_PHASE;
    COG1ASD1 = CFG_COG_1_ASD1;
    COG1ASD0 = CFG_COG_1_ASD0;
    COG1RIS0 = CFG_COG_1_REI0_NONE;
    COG1RIS1 = CFG_COG_1_REI1_PWM5;
    COG1FIS0 = CFG_COG_1_FEI0_NONE;

    if (!(statOpen) && !(statDigital)) COG1FIS1 = CFG_COG_1_FEI1_CLOOP;
    else COG1FIS1 = CFG_COG_1_FEI1_OLOOP;

    COG1RSIM0 = CFG_COG_1_REM0_NONE;
    COG1RSIM1 = CFG_COG_1_REM1_PWM5;
    COG1FSIM0 = CFG_COG_1_FEM0_NONE;
    COG1FSIM1 = CFG_COG_1_FEM1_LOOP;
    COG1CON1 = CFG_COG_1_CON1;
    COG1CON0 = CFG_COG_1_CON0;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Setup CLC2; should be done after COG1 config.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC2SEL0 = CFG_CLC_2_SEL_0;
    CLC2SEL1 = CFG_CLC_2_SEL_1;
    CLC2SEL2 = CFG_CLC_2_SEL_2;
    CLC2SEL3 = CFG_CLC_2_SEL_3;
    CLC2GLS0 = CFG_CLC_2_GLS_0;
    CLC2GLS1 = CFG_CLC_2_GLS_1;
    CLC2GLS2 = CFG_CLC_2_GLS_2;
    CLC2GLS3 = CFG_CLC_2_GLS_3;
    CLC2POL = CFG_CLC_2_POL;
    CLC2CON = CFG_CLC_2_CON;
    LC2EN = TRUE;                           //Enable


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure current sense S&H amp (OPA2)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OPA2CON = CFG_OPA_2_C;
    OPA2ORS = CFG_OPA_2_O;
    OPA2NCHS = CFG_OPA_2_N;
    OPA2PCHS = CFG_OPA_2_P;
    OPA2EN = TRUE;                          //Enable


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure comparator 5
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM5CON0 = CFG_CM_5_C0;
    CM5CON1 = CFG_CM_5_C1;
    CM5NSEL = CFG_CM_5_N;
    CM5PSEL = CFG_CM_5_P;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure Programmable Ramp Generator 0
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PRG3CON0 = CFG_PRG_3_C0;
    PRG3CON1 = CFG_PRG_3_C1;
    PRG3INS = CFG_PRG_3_VS;
    PRG3CON2 = CFG_PRG_3_C2;
    PRG3RTSS = CFG_PRG_3_RS;
    PRG3FTSS = CFG_PRG_3_FS;

    //Wait for ramp to stabilize and start
    while (!(RG3RDY));
    RG3GO = TRUE;


    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Enable interrupts

    TMR2IE = TRUE;                      //Enable timer2 interrupt
    T2ON = TRUE;                        //Start caretaker

    return;
} //End function Initialize



//End of file